/*
 * ip_vs_nfct.c:	Netfilter connection tracking support for IPVS
 *
 * Portions Copyright (C) 2001-2002
 * Antefacto Ltd, 181 Parnell St, Dublin 1, Ireland.
 *
 * Portions Copyright (C) 2003-2008
 * Julian Anastasov
 *
 *
 * This code is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * Authors:
 * Ben North <ben@redfrontdoor.org>
 * Julian Anastasov <ja@ssi.bg>		Reorganize and sync with latest kernels
 *
 *
 * Current status:
 *
 * - provide conntrack confirmation for new and related connections, by
 * this way we can see their proper conntrack state in all hooks
 * - support for all forwarding methods, not only NAT
 * - FTP support (NAT), ability to support other NAT apps with expectations
 * - to correctly create expectations for related NAT connections the proper
 * NF conntrack support must be already installed, eg. ip_vs_ftp requires
 * nf_conntrack_ftp for the same ports
 *
 */

#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/compiler.h>
#include <linux/vmalloc.h>
#include <linux/skbuff.h>
#include <net/ip.h>
#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>
#include <net/ip_vs.h>


EXPORT_SYMBOL(ip_vs_nfct_expect_related);


#define FMT_TUPLE	"%u.%u.%u.%u:%u->%u.%u.%u.%u:%u/%u"
#define ARG_TUPLE(t)	NIPQUAD((t)->src.u3.ip), ntohs((t)->src.u.all), \
			NIPQUAD((t)->dst.u3.ip), ntohs((t)->dst.u.all), \
			(t)->dst.protonum

#define FMT_CONN	"%u.%u.%u.%u:%u->%u.%u.%u.%u:%u->%u.%u.%u.%u:%u/%u:%u"
#define ARG_CONN(c)	NIPQUAD((c)->caddr), ntohs((c)->cport), \
			NIPQUAD((c)->vaddr), ntohs((c)->vport), \
			NIPQUAD((c)->daddr), ntohs((c)->dport), \
			(c)->protocol, (c)->state

/* Returns boolean and skb is freed on failure */
static int __ip_vs_nfct_confirm(struct sk_buff *skb, struct ip_vs_conn *cp,
				unsigned int hooknum)
{
	/*
	 * The assumptions:
	 * - the nfct is !NULL and is not confirmed
	 * - we are called before any mangle
	 */

	struct iphdr *iph = ip_hdr(skb);
	struct nf_conn *ct = (struct nf_conn *) skb->nfct;
	struct nf_conntrack_tuple new_reply;
	int ret = NF_DROP;
	__be16 _ports[2], *pptr;
#ifdef CONFIG_IP_VS_DEBUG
	struct nf_conntrack_tuple *orig_tup =
		&ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple;
	struct nf_conntrack_tuple *orig_rep =
		&ct->tuplehash[IP_CT_DIR_REPLY].tuple;
#endif
#ifdef CONFIG_NF_NAT_NEEDED
	int initialized = !!(ct->status & IPS_NAT_DONE_MASK);
#else
	int initialized = 0;
#endif

	IP_VS_DBG(7, "%s: ct=%p, init=%d, tuples=" FMT_TUPLE ", " FMT_TUPLE
		", cp=" FMT_CONN "\n",
		__FUNCTION__, ct, initialized,
		ARG_TUPLE(orig_tup), ARG_TUPLE(orig_rep), ARG_CONN(cp));

#ifdef CONFIG_NF_NAT_NEEDED
	/*
	 * This is really bad, may be we are trying to alter DNAT conn?
	 * This is not supported, avoid the confirmation.
	 */
	if (initialized && ct->status & IPS_NAT_MASK) {
#ifdef CONFIG_IP_VS_DEBUG
		IP_VS_DBG(7, "%s: ct=%p, status=0x%lX, init=%d\n",
			__FUNCTION__, ct, ct->status, initialized);
#endif
		return 1;
	}
#endif

	if (IP_VS_FWD_METHOD(cp) != IP_VS_CONN_F_MASQ || NF_INET_FORWARD == hooknum)
		goto confirm;

	/*
	 * Alter reply only for IP_VS_CONN_F_MASQ in outin direction.
	 * For related connections in inout direction it is done in
	 * expectfn callback.
	 */

	pptr = skb_header_pointer(skb, ip_hdrlen(skb),
				  sizeof(_ports), _ports);
	if (!pptr)
		goto out;

	new_reply = (struct nf_conntrack_tuple) {
		.dst = { .protonum = iph->protocol, .dir = IP_CT_DIR_REPLY }};

	new_reply.src.u3.ip = cp->daddr;
	new_reply.src.u.tcp.port = cp->dport;
	new_reply.src.l3num = PF_INET;
	new_reply.dst.u3.ip = iph->saddr;
	new_reply.dst.u.tcp.port = pptr[0];

	nf_conntrack_alter_reply(ct, &new_reply);

	IP_VS_DBG(7, "%s: ct=%p, init=%d, orig=" FMT_TUPLE
		", new_reply=" FMT_TUPLE " => alter_reply\n",
		__FUNCTION__, ct, initialized,
		ARG_TUPLE(orig_tup), ARG_TUPLE(&new_reply));

	/*
	 * No need to rehash NAT info because we don't change source
	 * address in original direction
	 */

confirm:

	ret = __nf_conntrack_confirm(skb);

	if (ret != NF_STOLEN) {
		IP_VS_DBG(7, "%s: ct=%p, init=%d, orig=" FMT_TUPLE " => confirm ret=%d\n",
			__FUNCTION__, ct, initialized, ARG_TUPLE(orig_tup), ret);
	}

	if (ret != NF_ACCEPT)
		goto out;
	return 1;

out:
	if (ret != NF_STOLEN)
		kfree_skb(skb);
	return 0;
}

/*
 * Confirm (and optionally alter) the conntrack entry if needed
 * because the IPVS packets do not reach ipv4_confirm.
 */
int ip_vs_nfct_confirm(struct sk_buff *skb, struct ip_vs_conn *cp,
		       unsigned int hooknum)
{
	struct iphdr *iph = ip_hdr(skb);
	struct nf_conn *ct = (struct nf_conn *) skb->nfct;

	/* By the time we're sending the packet out the other
	 * side, there should be a confirmed Netfilter CT entry
	 * for this connection.  This may not be the case,
	 * however, if it's a brand new connection, or if the NF
	 * entry has timed out before ours has.  Either way, if
	 * the NF CT entry is unconfirmed, confirm it, and deal
	 * with reply tuple mangling at the same time.
	 */

	/* We only deal with TCP or UDP packets */
	if (iph->protocol != IPPROTO_TCP && iph->protocol != IPPROTO_UDP)
		return 1;

	if (IP_VS_FWD_METHOD(cp) != IP_VS_CONN_F_MASQ) {
		/*
		 * Do not be surprised if non-NAT conntracks stay in SYN_SENT
		 * state, may be the replies from the real server go
		 * directly to client. In any case, keep them in REPLIED
		 * state (ESTABLISHED).
		 */
		if (iph->protocol != IPPROTO_TCP ||
		    IP_VS_TCP_S_ESTABLISHED == cp->state) {
			set_bit(IPS_SEEN_REPLY_BIT, &ct->status);
		}
	}

	/*
	 * We assume the reused connections do not change their rip:rport
	 * and we do not need to alter their conntrack reply
	 */
	return __ip_vs_nfct_confirm(skb, cp, hooknum);
}

/*
 * We are called from init_conntrack() as expectfn handler
 */

static void ip_vs_nfct_expect_callback(struct nf_conn *ct,
	struct nf_conntrack_expect *exp)
{
	struct nf_conntrack_tuple *orig, new_reply;
	struct ip_vs_conn *cp;

	if (exp->tuple.src.l3num != PF_INET)
		return;

	/* 
	 * - We assume that no NF locks are held before this callback
	 * - ip_vs_conn_out_get and ip_vs_conn_in_get should match their
	 * expectations even if they use wildcard values, now we provide
	 * the actual values from the newly created original conntrack direction
	 * - the conntrack is confirmed when packet reaches IPVS hooks
	 */

	/* RS->CLIENT */
	orig = &ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple;
	cp = ip_vs_conn_out_get(orig->dst.protonum,
				orig->src.u3.ip, orig->src.u.tcp.port,
				orig->dst.u3.ip, orig->dst.u.tcp.port);
	if (cp) {
		/* Change reply CLIENT->RS to CLIENT->VS */
		new_reply = ct->tuplehash[IP_CT_DIR_REPLY].tuple;
		IP_VS_DBG(7, "%s: ct=%p, status=0x%lX, tuples=" FMT_TUPLE ", " FMT_TUPLE
			", found inout cp=" FMT_CONN "\n",
			__FUNCTION__, ct, ct->status,
			ARG_TUPLE(orig), ARG_TUPLE(&new_reply),
			ARG_CONN(cp));
		new_reply.dst.u3.ip = cp->vaddr;
		new_reply.dst.u.tcp.port = cp->vport;
		IP_VS_DBG(7, "%s: ct=%p, new tuples=" FMT_TUPLE ", " FMT_TUPLE
			", inout cp=" FMT_CONN "\n",
			__FUNCTION__, ct,
			ARG_TUPLE(orig), ARG_TUPLE(&new_reply),
			ARG_CONN(cp));
		goto alter;
	}

	/* CLIENT->VS */
	cp = ip_vs_conn_in_get(orig->dst.protonum,
				orig->src.u3.ip, orig->src.u.tcp.port,
				orig->dst.u3.ip, orig->dst.u.tcp.port);
	if (cp) {
		/* Change reply VS->CLIENT to RS->CLIENT */
		new_reply = ct->tuplehash[IP_CT_DIR_REPLY].tuple;
		IP_VS_DBG(7, "%s: ct=%p, status=0x%lX, tuples=" FMT_TUPLE ", " FMT_TUPLE
			", found outin cp=" FMT_CONN "\n",
			__FUNCTION__, ct, ct->status,
			ARG_TUPLE(orig), ARG_TUPLE(&new_reply),
			ARG_CONN(cp));
		new_reply.src.u3.ip = cp->daddr;
		new_reply.src.u.tcp.port = cp->dport;
		IP_VS_DBG(7, "%s: ct=%p, new tuples=" FMT_TUPLE ", " FMT_TUPLE
			", outin cp=" FMT_CONN "\n",
			__FUNCTION__, ct,
			ARG_TUPLE(orig), ARG_TUPLE(&new_reply),
			ARG_CONN(cp));
		goto alter;
	}
	IP_VS_DBG(7, "%s: ct=%p, status=0x%lX, tuple=" FMT_TUPLE " - unknown expect\n",
		__FUNCTION__, ct, ct->status, ARG_TUPLE(orig));
	return;

alter:

	/* Never alter conntrack for non-NAT conns */
	if (IP_VS_FWD_METHOD(cp) == IP_VS_CONN_F_MASQ)
		nf_conntrack_alter_reply(ct, &new_reply);
	ip_vs_conn_put(cp);
	return;
}

/*
 * Create NF conntrack expectation with wildcard (optional) source port.
 * Then the default callback function will alter the reply and will confirm
 * the conntrack entry when the first packet comes.
 */
void ip_vs_nfct_expect_related(struct sk_buff *skb, struct ip_vs_conn *cp,
			       __be16 port, __u16 proto, int from_rs)
{
	struct nf_conn *ct = (struct nf_conn *) skb->nfct;
	struct nf_conntrack_expect *e;

	if (!sysctl_ip_vs_conntrack)
		return;

	if (!ct) {
		IP_VS_DBG(7, "%s: ct=%p for cp=" FMT_CONN "\n",
			__FUNCTION__, ct, ARG_CONN(cp));
		return;
	}

	if (!(e = nf_ct_expect_alloc(ct)))
		return;

	e->expectfn			= ip_vs_nfct_expect_callback;
	e->helper			= NULL;
	e->flags			= 0;
	e->class			= NF_CT_EXPECT_CLASS_DEFAULT;
	memset(&e->tuple, 0, sizeof(e->tuple));
	e->tuple.src.u.tcp.port		= port;
	e->tuple.src.l3num		= PF_INET;
	e->tuple.dst.protonum		= proto;
	memset(&e->mask, 0, sizeof(e->mask));
	e->mask.src.u3.ip		= 0xffffffff;
	e->mask.src.u.all		= port? 0xffff : 0;

	if (from_rs) {
		e->tuple.src.u3.ip = cp->daddr;
		e->tuple.dst.u3.ip = cp->caddr;
		e->tuple.dst.u.tcp.port = cp->cport;
	} else {
		e->tuple.src.u3.ip = cp->caddr;
		e->tuple.dst.u3.ip = cp->vaddr;
		e->tuple.dst.u.tcp.port = cp->vport;
	}

	IP_VS_DBG(7, "%s: ct=%p, expect tuple=" FMT_TUPLE "\n",
		__FUNCTION__, ct, ARG_TUPLE(&e->tuple));
	nf_ct_expect_related(e);
	nf_ct_expect_put(e);
}

/*
 * Our connection was terminated, try to drop the conntrack immediately
 */
void ip_vs_nfct_conn_drop(struct ip_vs_conn *cp)
{
	struct nf_conntrack_tuple_hash *h;
	struct nf_conn *ct;
	struct nf_conntrack_tuple tuple;

	if (!cp->cport)
		return;

	tuple = (struct nf_conntrack_tuple) {
		.dst = { .protonum = cp->protocol, .dir = IP_CT_DIR_ORIGINAL } };
	tuple.src.u3.ip = cp->caddr;
	tuple.src.u.all = cp->cport;
	tuple.src.l3num = PF_INET;
	tuple.dst.u3.ip = cp->vaddr;
	tuple.dst.u.all = cp->vport;

	IP_VS_DBG(7, "%s: dropping conntrack with tuple=" FMT_TUPLE
		" for conn " FMT_CONN "\n",
		__FUNCTION__, ARG_TUPLE(&tuple), ARG_CONN(cp));

	h = nf_conntrack_find_get(&tuple);
	if (h) {
		ct = nf_ct_tuplehash_to_ctrack(h);
		if (del_timer(&ct->timeout)) {
			IP_VS_DBG(7, "%s: ct=%p, deleted conntrack timer for tuple="
				FMT_TUPLE "\n",
				__FUNCTION__, ct, ARG_TUPLE(&tuple));
			if (ct->timeout.function)
				ct->timeout.function(ct->timeout.data);
		} else {
			IP_VS_DBG(7, "%s: ct=%p, no conntrack timer for tuple="
				FMT_TUPLE "\n",
				__FUNCTION__, ct, ARG_TUPLE(&tuple));
		}
		nf_ct_put(ct);
	} else {
		IP_VS_DBG(7, "%s: no conntrack for tuple=" FMT_TUPLE "\n",
			__FUNCTION__, ARG_TUPLE(&tuple));
	}
}

